import uuid
from enum import Enum
from typing import Dict, Any, List, Optional
from pydantic import BaseModel, model_validator


class PlotLibrary(str, Enum):
    PLT = "PLT"
    SHAP = "SHAP"


class PlotType(str, Enum):
    BAR = "BAR"
    WATERFALL = "WATERFALL"
    PLOT = "PLOT"


class IndividualPlot(BaseModel):
    plot_type: PlotType
    kwargs: Optional[Dict] = {}
    args: Optional[List] = []
    xlabel: Optional[str] = None
    ylabel: Optional[str] = None
    title: Optional[str] = None

    @model_validator(mode="before")
    @classmethod
    def backwards_compatibility_plot_data(cls, data: Any) -> Any:
        if isinstance(data, dict):
            if "plot_data" in data:
                plot_data = data.pop("plot_data")
                data["kwargs"] = plot_data
        return data

    def model_dump(self, **kwargs: Any) -> Dict[str, Any]:
        # configure parent method
        kwargs.setdefault("exclude_none", True)
        kwargs.setdefault("exclude_defaults", True)
        return super().model_dump(**kwargs)

    def plot(self, plot_library: PlotLibrary, fig: "matplotlib.pyplot.Figure") -> None:
        import shap
        import numpy as np

        shap_plot_function_mapping = {"WATERFALL": shap.plots.waterfall}
        plt_plot_function_mapping = {"PLOT": fig.gca().plot, "BAR": fig.gca().bar}

        plot_type = self.plot_type
        kwargs = self.kwargs
        args = self.args

        if plot_library == "SHAP":
            shap_values = shap._explanation.Explanation(
                values=np.array(kwargs["values"]),
                base_values=kwargs["base_value"],
                data=np.array(kwargs["data"]),
                feature_names=kwargs["feature_names"],
            )

            shap_plot_function_mapping[plot_type](shap_values, show=False)

        elif plot_library == "PLT":
            from matplotlib.ticker import ScalarFormatter

            if "color" not in kwargs:
                kwargs["color"] = "#0eab9c"

            plt_plot_function_mapping[plot_type](*args, **kwargs)

            fig.gca().spines["top"].set_visible(False)
            fig.gca().spines["right"].set_visible(False)

            fig.gca().spines["left"].set_color("#515353")
            fig.gca().spines["bottom"].set_color("#515353")

            # Create a ScalarFormatter object
            formatter = ScalarFormatter(useOffset=False)
            formatter.set_scientific(False)
            formatter.set_useOffset(False)

            # Apply the formatter to the x-axis and y-axis
            fig.gca().xaxis.set_major_formatter(formatter)
            fig.gca().yaxis.set_major_formatter(formatter)

            for axis in fig.axes:
                if self.title:
                    axis.set_title(self.title)
                if self.xlabel:
                    axis.set_xlabel(self.xlabel, color="#515353")
                if self.ylabel:
                    axis.set_ylabel(self.ylabel, color="#515353")

            fig.gca().tick_params(axis="x", colors="#515353")
            fig.gca().tick_params(axis="y", colors="#515353")

        else:
            raise ValueError(f"Unknown plot library {plot_library}")


class PlotData(BaseModel):
    plot_name: str
    plot_library: PlotLibrary
    plot_content: List[IndividualPlot]

    def create_plot(self) -> "matplotlib.pyplot.Figure":
        from matplotlib import pyplot as plt

        plot_library = self.plot_library
        plot_content = self.plot_content

        fig = plt.figure(num=str(uuid.uuid4()))
        for individual_plot in plot_content:
            individual_plot.plot(plot_library=plot_library, fig=fig)

        return fig



from datetime import datetime
from typing import List, Dict, Optional, Union, TYPE_CHECKING
from uuid import UUID

import requests
import structlog
from requests.auth import HTTPBasicAuth

if TYPE_CHECKING:
    import pandas as pd

from tarandm_analytics.base_class import TaranDMAnalytics

logger = structlog.get_logger(__name__)


class EvaluateAttributes(TaranDMAnalytics):
    """
    Attribute evaluator is designed to evaluate attributes from selected AttributeClass. User can select time range and
    business case. Matched requests from TaranDM database will be processed and all attributes from selected attribute
    classes will be evaluated into pandas dataframe. Such dataset can be used for instance for predictive model
    development.
    """

    def __init__(self, endpoint_url: Optional[str], username: Optional[str], password: Optional[str]):
        super().__init__(endpoint_url=endpoint_url, username=username, password=password)
        self.authorization: HTTPBasicAuth = HTTPBasicAuth(self.username, self.password)
        self.last_attribute_extractor_id: Optional[Dict[str, Union[str, datetime]]] = None
        self.attribute_extractor_ids: List[Dict[str, Union[str, datetime]]] = []

    def check_evaluation_progress(self, attribute_extractor_id: Optional[str] = None) -> str:
        """
        Once `evaluate` method is triggered, attribute extraction process is created with assigned id. This method
        can be used to check if the process has finished.

        :param attribute_extractor_id: Attribute extraction process id.
        :return:
        """
        if attribute_extractor_id is None:
            if self.last_attribute_extractor_id is None:
                return (
                    "Attribute extraction was not triggered yet. You can also provide attribute_extractor_id as a "
                    "parameter to extract data from past extraction process."
                )
            attribute_extractor_id = self.last_attribute_extractor_id["id"]

        url = self.endpoint_url + "analytics/check_attributes_evaluator_progress"
        request_data = {"process_id": attribute_extractor_id}

        response = requests.post(url=url, json=request_data, auth=self.authorization)
        if response.status_code == 200:
            logger.info("Attribute evaluation progress status checked.")

        return response.json()["status"]

    def evaluate(
        self,
        attribute_classes: List[str],
        input_data_class: str,
        business_case: str,
        repository: str,
        git_user_name: str,
        git_user_email: str,
        git_user_token: str,
        date_from: Optional[str] = None,
        date_to: Optional[str] = None,
        decision_ids: Optional[List[str]] = None,
        orig_git_branch: Optional[str] = "develop-teambox",
    ) -> None:
        """
        Run attribute extraction process.

        :param attribute_classes: List of Attribute classes to be evaluated. For available classes, run
        `get_attribute_classes` method.
        :param input_data_class: Expected input data class.
        :param business_case: Process requests of given business case only.
        :param repository: Git repository where strategies are located.
        :param git_user_name: Git authorization - user name.
        :param git_user_email: Git authorization - user email.
        :param git_user_token: Git authorization - personal access token.
        :param date_from: Only requests after this date will be considered. If not provided, no limitation is applied.
        :param date_to: Only requests before this date will be considered. If not provided, no limitation is applied.
        :param decision_ids: List of decision ids to be processed. If not provided, decision ids are fetched based on
        `business_case`, `date_from` and `date_to`.
        :param orig_git_branch: Git branch to be used. Different branches can contain different attribute definitions.
        :return:
        """
        url = self.endpoint_url + "analytics/attributes_evaluator"

        date_from = date_from or "1900-01-01"
        date_to = date_to or "2100-12-31"

        request_data = {
            "date_from": date_from,
            "date_to": date_to,
            "decision_ids": [UUID(did) for did in decision_ids] if decision_ids is not None else None,
            "attribute_classes": attribute_classes,
            "input_data_class": input_data_class,
            "business_case": business_case,
        }

        repository_config = {
            "class": "tarandm_utils.model.repository_config.project.ProjectRepositoryConfig",
            "class_version": "a492a7ba",
            "readonly": "false",
            "source": "Attribute evaluator",
            "customer_version": "local",
            "repository": repository,
            "branch": f"{orig_git_branch}-attribute-evaluator",
            "original_branch": orig_git_branch,
            "user_name": git_user_name,
            "user_email": git_user_email,
            "user_token": git_user_token,
        }

        headers = {"tarandm-repository_config": str(repository_config).replace("'", '"')}

        response = requests.post(url=url, params=request_data, headers=headers, auth=self.authorization)

        if response.status_code == 200:
            logger.info("Attributes were evaluated and stored to DB.")
        else:
            logger.error(f"Was not able to evaluate attributes. Status code {response.status_code}.")

        self.last_attribute_extractor_id = {"id": response.json().get("simulation_id"), "created": datetime.now()}
        self.attribute_extractor_ids.append(self.last_attribute_extractor_id)

    def get_attribute_classes(self) -> Optional[Dict[str, List[str]]]:
        """
        Get available attribute classes.

        :return: Dictionary with attribute classes as keys and list of class attributes as values.
        """
        url = self.endpoint_url + "strategies/attribute_classes"

        response = requests.post(url=url, auth=self.authorization)

        if response.status_code != 200:
            logger.error(f"Failed to fetch list of attribute classes: {response.status_code}")
            return None

        result = {}
        for attr_class_source, attr_classes in response.json().items():
            for attr_class in attr_classes:
                result[attr_class["class_name"]] = list(attr_class["attributes"].keys())

        return result

    def get_business_cases(self, selector_name: str = "demo") -> Optional[Dict[str, Dict[str, str]]]:
        """
        Get available business cases.

        :param selector_name: Selector name.
        :return: Dictionary with available business cases as keys. Values contain input data class and audience for
        which the business case is defined.
        """
        url = self.endpoint_url + "strategies/selectors"

        response = requests.post(
            url=url, json={"action": "get", "params": {"name": selector_name}}, auth=self.authorization
        )

        if response.status_code != 200:
            logger.error(f"Failed to fetch list of attribute classes: {response.status_code}")
            return None

        result = {}
        for business_case in response.json()["business_cases"]:
            result[business_case["name"]] = {
                "input_class": business_case["input_class"],
                "audiences": [audience["query"] for audience in business_case["audiences"]],
            }

        return result

    def fetch_data_from_db(self, attribute_extractor_id: Optional[str] = None) -> Optional["pd.DataFrame"]:
        """
        Once attribute extraction process finishes, dataset can be extracted using this method.

        :param attribute_extractor_id: Attribute extraction process id.
        :return: DataFrame with evaluated attributes.
        """
        status = self.check_evaluation_progress()

        if status == "RUNNING":
            logger.info("Attribute evaluation still running.")
            return None
        elif status != "FINISHED":
            logger.warning(f"Attribute evaluation status: {status}.")
            return None

        logger.info("Attribute evaluation finished. Data fetching stated.")

        if attribute_extractor_id is None:
            if self.last_attribute_extractor_id is None:
                logger.info(
                    "Attribute extraction was not triggered yet. You can also provide attribute_extractor_id "
                    "as a parameter to extract data from past extraction process."
                )
                return None
            attribute_extractor_id = self.last_attribute_extractor_id["id"]

        url = self.endpoint_url + "analytics/get_attributes_evaluator_data"
        request_data = {"process_id": attribute_extractor_id}

        response = requests.post(url=url, json=request_data, auth=self.authorization)
        if response.status_code == 200:
            logger.info("Attribute evaluation progress status checked.")

        import pandas as pd

        df = pd.DataFrame()
        for decision_id, db_row in response.json().items():
            pd_row = {
                "decision_id": decision_id,
            }

            for attr_name, content in db_row.items():
                name = content["name"]
                value = content["value"]

                if isinstance(value, dict) and "amount" in value.keys():
                    amount = value.get("amount")
                    currency = value.get("currency", "")
                    pd_row[name] = f"{amount}{currency}"
                else:
                    pd_row[name] = value

            df = pd.concat([df, pd.DataFrame(pd.Series(pd_row)).T])

        return df



import io
import os
import shutil
import stat
import urllib
from datetime import datetime
from pathlib import Path
from typing import List, Dict, Any, Optional

import gitlab
import structlog
from furl import furl
from git import Repo

from tarandm_analytics.export_predictive_model.create_predictive_model import ExportPredictiveModel

logger = structlog.get_logger(__name__)


def remove_readonly(func, path, _):
    """Clear the readonly bit and reattempt the removal"""
    os.chmod(path, stat.S_IWRITE)
    func(path)


def delete_folder(folder):
    # Check if the folder exists and delete it if it does
    if os.path.exists(folder):
        logger.info(f"Folder {folder} already exists, deleting.")
        shutil.rmtree(folder, onerror=remove_readonly)


def create_folder(folder):
    os.makedirs(folder)
    logger.info(f'Creating folder"{folder}".')


def create_merge_request(
    gitlab_url: str,
    repository_url: str,
    environment: str,
    git_user_name: str,
    git_user_token: str,
    cloned_repo_dir: str,
    request_data: Dict[str, Any],
    images: Optional[List[Dict[str, Any]]],
    epm: ExportPredictiveModel,
) -> None:
    # Get the current time
    current_time = datetime.now()

    # Format the time as a string
    time_str = current_time.strftime("%Y%m%d%H%M%S")

    new_model_name = request_data["model_name"]
    teambox_branch_name = f"{environment}-teambox"
    new_branch_name = f"{environment}-feature/{git_user_name}/upload_new_model_{time_str}"
    path_to_models = Path("repository/strategies/models")
    repo_full_path = "tarandm/strategies"

    # Prepare the folder
    delete_folder(cloned_repo_dir)
    create_folder(cloned_repo_dir)

    # prepare furl url
    repo_url = furl(repository_url)
    repo_url.username = "__token__"
    repo_url.password = git_user_token

    # clone strategies repo
    repo = Repo.clone_from(
        urllib.parse.unquote(str(repo_url)),
        cloned_repo_dir,
        branch=teambox_branch_name,
    )
    logger.info(f"Cloned strategies repository into {cloned_repo_dir}")

    # Create a new branch from the current branch
    repo.create_head(new_branch_name)
    logger.info(f"New branch '{new_branch_name}' created.")

    # Switch to the new branch (optional)
    repo.git.checkout(new_branch_name)
    logger.info(f"Switched to branch '{new_branch_name}'")

    # get necessary files
    extended_model_yaml, external_model_json, external_model_pmml, images = epm._build_predictive_model(
        request_data=request_data, images=images
    )

    # prepare new model folder
    new_model_folder = Path(cloned_repo_dir) / path_to_models / Path(new_model_name)

    if os.path.exists(new_model_folder):
        model_already_existed = True
    else:
        model_already_existed = False

    delete_folder(new_model_folder)
    create_folder(new_model_folder)

    # Write extended_model_yaml to file
    if extended_model_yaml:
        with io.StringIO(extended_model_yaml) as file:
            filepath = new_model_folder / Path("extended_model.yaml")
            with open(filepath, "w") as outfile:
                outfile.write(file.getvalue())
            logger.info(f"Wrote extended_model_yaml to {filepath}")

    # Write external_model_json to file
    if external_model_json:
        with io.StringIO(external_model_json) as file:
            filepath = new_model_folder / Path("external_model.json")
            with open(filepath, "w") as outfile:
                outfile.write(file.getvalue())
            logger.info(f"Wrote external_model_json to {filepath}")

    # Write external_model_pmml to file
    if external_model_pmml:
        with io.StringIO(external_model_pmml) as file:
            filepath = new_model_folder / Path("external_model.pmml")
            with open(filepath, "w") as outfile:
                outfile.write(file.getvalue())
            logger.info(f"Wrote external_model_pmml to {filepath}")

    # Write images to files
    for image_info in images:
        filename = image_info["filename"]
        image_content = image_info["image"]
        image_content.seek(0)
        filepath = new_model_folder / Path(filename)
        with open(filepath, "wb") as outfile:
            outfile.write(image_content.read())
        logger.info(f"Wrote image content to {filepath}")

    logger.info(f"All files have been written to the folder '{new_model_folder}'.")

    # Git add, commit and push
    subfolder_path = path_to_models / Path(new_model_name)
    repo.git.add(subfolder_path)

    if model_already_existed:
        message = f"Update model {new_model_name}."
    else:
        message = f"Create model {new_model_name}"

    repo.index.commit(message=message)
    repo.git.push("origin", new_branch_name)
    logger.info("Changes commited and pushed to GitLab")

    # initialize GitLab
    gl = gitlab.Gitlab(gitlab_url, private_token=git_user_token)
    # Get the project
    project = gl.projects.get(repo_full_path)

    # Create the merge request
    merge_request = project.mergerequests.create(
        {
            "source_branch": new_branch_name,
            "target_branch": teambox_branch_name,
            "title": message,
            "description": message,
        }
    )

    logger.info(f"Merge request created: {merge_request.web_url}")

    delete_folder(cloned_repo_dir)



from typing import Union


def get_number_formatting(number: Union[float, int]) -> str:
    import numpy as np

    if number == -np.inf:
        return "-inf"
    elif isinstance(number, int):
        return str(number)
    elif number * 100 > 1:
        return "{:.3f}".format(number)
    else:
        return "{:.2E}".format(number)



from typing import Optional

import requests
import structlog
from requests.auth import HTTPBasicAuth

logger = structlog.get_logger(__name__)


class TaranDMAnalytics:
    def __init__(
        self,
        endpoint_url: Optional[str],
        username: Optional[str],
        password: Optional[str],
        skip_endpoint_validation: bool = False,
    ) -> None:
        if endpoint_url is not None:
            self.endpoint_url = endpoint_url + ("" if endpoint_url.endswith("/") else "/")
        self.username = username
        self.password = password

        # endpoint validation might be skipped if only methods that do not use endpoint are to be used
        if not skip_endpoint_validation:
            self.validate_url()

    def validate_url(self) -> None:
        url = self.endpoint_url + "info"
        response = requests.get(url=url, auth=HTTPBasicAuth(self.username, self.password))

        if response.status_code == 200:
            logger.info(f"Connection to {self.endpoint_url} was established.")
        elif response.status_code == 401:
            logger.info(
                f"Connection to {self.endpoint_url} cannot be established. Endpoint error message: " f"{response.text}"
            )
